@@ -482,7 +482,6 @@ class objc_property:
         return (objc_property_attribute_t * len(attrs))(*attrs)
 
     def class_register(self, class_ptr, attr_name):
-
         ivar_name = "_" + attr_name
 
         add_ivar(class_ptr, ivar_name, self.vartype)
@@ -519,7 +518,6 @@ class objc_property:
             return value
 
         def _objc_setter(objc_self, _cmd, new_value):
-
             if self._is_py_object and self.weak:
                 # Don't store the object itself but only a Python weakref.
                 new_value = weakref.ref(new_value)
@@ -575,7 +573,6 @@ class objc_property:
         libobjc.class_addProperty(class_ptr, ensure_bytes(attr_name), attrs, len(attrs))
 
     def dealloc_callback(self, objc_self, attr_name):
-
         ivar_name = "_" + attr_name
 
         if self._ivar_weak:
@@ -891,14 +888,20 @@ class ObjCInstance:
                 #
                 # Refs #249.
                 if cls == ObjCInstance or isinstance(cls, ObjCInstance):
-                    cached_class_name = ensure_bytes(cached.objc_class.name)
+                    cached_class_name = cached.objc_class.name
                     current_class_name = libobjc.class_getName(
                         libobjc.object_getClass(object_ptr)
-                    )
-                    if cached_class_name != current_class_name:
-                        # There has been a cache hit, but the object is a different class.
-                        # Purge the cache for this address, and treat it as a cache miss.
-                        del cls._cached_objects[object_ptr.value]
+                    ).decode("utf-8")
+                    if (
+                        current_class_name != cached_class_name
+                        and not current_class_name.endswith(f"_{cached_class_name}")
+                    ):
+                        # There has been a cache hit, but the object is a
+                        # different class, treat this as a cache miss. We don't
+                        # *just* look for an *exact* class name match, because
+                        # some Cocoa/UIKit classes undergo a class name change
+                        # between `alloc()` and `init()` (e.g., `NSWindow`
+                        # becomes `NSKVONotifying_NSWindow`). Refs #257.
                         raise KeyError(object_ptr.value)
 
                 return cached
@@ -1280,7 +1283,6 @@ class ObjCClass(ObjCInstance, type):
         user_dealloc = attrs.get("dealloc", None)
 
         def _new_delloc(objc_self, _cmd):
-
             # Invoke user-defined dealloc.
             if user_dealloc:
                 user_dealloc(objc_self, _cmd)
@@ -1994,7 +1996,6 @@ try:
 except NameError:
 
     class WrappedPyObject(NSObject):
-
         wrapped_pointer = objc_ivar(c_void_p)
 
         @objc_rawmethod
